@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 ##############################################################################
-# Copyright (c) 2002 Zope Corporation and Contributors.
+# Copyright (c) 2002 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -12,8 +12,6 @@
 #
 ##############################################################################
 """Schema Fields
-
-$Id: _field.py 107713 2010-01-05 17:10:46Z chrism $
 """
 
 __docformat__ = 'restructuredtext'
@@ -23,7 +21,7 @@ import decimal
 from datetime import datetime, date, timedelta, time
 from zope.event import notify
 
-from zope.interface import classImplements, implements
+from zope.interface import classImplements, implements, Interface
 from zope.interface.interfaces import IInterface, IMethod
 
 from zope.schema.interfaces import IField
@@ -168,7 +166,7 @@ class Float(Orderable, Field):
         >>> f.fromUnicode("1.25")
         1.25
         >>> f.fromUnicode("1.25.6") #doctest: +IGNORE_EXCEPTION_DETAIL
-        Traceback (most recent call last): 
+        Traceback (most recent call last):
         ...
         ValueError: invalid literal for float(): 1.25.6
         """
@@ -457,18 +455,37 @@ class FrozenSet(AbstractCollection):
 def _validate_fields(schema, value, errors=None):
     if errors is None:
         errors = []
-    for name in schema.names(all=True):
-        if not IMethod.providedBy(schema[name]):
-            try:
-                attribute = schema[name]
-                if IField.providedBy(attribute):
-                    # validate attributes that are fields
-                    attribute.validate(getattr(value, name))
-            except ValidationError, error:
-                errors.append(error)
-            except AttributeError, error:
-                # property for the given name is not implemented
-                errors.append(SchemaNotFullyImplemented(error))
+    # Interface can be used as schema property for Object fields that plan to
+    # hold values of any type.
+    # Because Interface does not include any Attribute, it is obviously not
+    # worth looping on its methods and filter them all out.
+    if schema is Interface:
+        return errors
+    # if `value` is part of a cyclic graph, we need to break the cycle to avoid
+    # infinite recursion.
+    if hasattr(value, '__schema_being_validated'):
+        return errors
+    # Mark the value as being validated.
+    value.__schema_being_validated = True
+    # (If we have gotten here, we know that `value` provides an interface
+    # other than zope.interface.Interface;
+    # iow, we can rely on the fact that it is an instance
+    # that supports attribute assignment.)
+    try:
+        for name in schema.names(all=True):
+            if not IMethod.providedBy(schema[name]):
+                try:
+                    attribute = schema[name]
+                    if IField.providedBy(attribute):
+                        # validate attributes that are fields
+                        attribute.validate(getattr(value, name))
+                except ValidationError, error:
+                    errors.append(error)
+                except AttributeError, error:
+                    # property for the given name is not implemented
+                    errors.append(SchemaNotFullyImplemented(error))
+    finally:
+        delattr(value, '__schema_being_validated')
     return errors
 
 
